@@ -1,5 +1,5 @@
 // -*- c++ -*-
-/* $Id: ustring.cc,v 1.7 2006/05/12 08:08:43 murrayc Exp $ */
+/* $Id: ustring.cc 652 2008-04-13 10:52:16Z murrayc $ */
 
 /* Copyright (C) 2002 The gtkmm Development Team
  *
@@ -22,14 +22,19 @@
 #include <glibmm/convert.h>
 #include <glibmm/error.h>
 #include <glibmm/utility.h>
+#include <glib.h>
 
 #include <algorithm>
 #include <iostream>
+#include <cstring>
 
+#include <config.h>
 #include <glibmmconfig.h>
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
+# include <stdexcept>
+#endif
 GLIBMM_USING_STD(find)
 
-
 namespace
 {
 
@@ -54,8 +59,8 @@ struct UnicharToUtf8
 // std::string will do that for us.
 
 // First overload: stop on '\0' character.
-//
-static ustring::size_type utf8_byte_offset(const char* str, ustring::size_type offset)
+static
+ustring::size_type utf8_byte_offset(const char* str, ustring::size_type offset)
 {
   if(offset == ustring::npos)
     return ustring::npos;
@@ -77,8 +82,8 @@ static ustring::size_type utf8_byte_offset(const char* str, ustring::size_type o
 }
 
 // Second overload: stop when reaching maxlen.
-//
-static ustring::size_type utf8_byte_offset(const char* str, ustring::size_type offset,
+static
+ustring::size_type utf8_byte_offset(const char* str, ustring::size_type offset,
                                     ustring::size_type maxlen)
 {
   if(offset == ustring::npos)
@@ -139,8 +144,8 @@ ustring::size_type utf8_char_offset(const std::string& str, ustring::size_type o
 
 // Helper to implement ustring::find_first_of() and find_first_not_of().
 // Returns the UTF-8 character offset, or ustring::npos if not found.
-//
-static ustring::size_type utf8_find_first_of(const std::string& str, ustring::size_type offset,
+static
+ustring::size_type utf8_find_first_of(const std::string& str, ustring::size_type offset,
                                       const char* utf8_match, long utf8_match_size,
                                       bool find_not_of)
 {
@@ -175,8 +180,8 @@ static ustring::size_type utf8_find_first_of(const std::string& str, ustring::si
 
 // Helper to implement ustring::find_last_of() and find_last_not_of().
 // Returns the UTF-8 character offset, or ustring::npos if not found.
-//
-static ustring::size_type utf8_find_last_of(const std::string& str, ustring::size_type offset,
+static
+ustring::size_type utf8_find_last_of(const std::string& str, ustring::size_type offset,
                                      const char* utf8_match, long utf8_match_size,
                                      bool find_not_of)
 {
@@ -197,7 +202,9 @@ static ustring::size_type utf8_find_last_of(const std::string& str, ustring::siz
   while(pstr > str_begin)
   {
     // Move to previous character.
-    do --pstr; while((*pstr & '\xC0') == '\x80');
+    do
+      --pstr;
+    while((static_cast<unsigned char>(*pstr) & 0xC0u) == 0x80);
 
     const gunichar *const pfound = std::find(match_begin, match_end, g_utf8_get_char(pstr));
 
@@ -220,36 +227,44 @@ namespace Glib
 const ustring::size_type ustring::npos = std::string::npos;
 #endif
 
-// We need our own version of g_utf8_get_char(), because the std::string
-// iterator is not necessarily a plain pointer (it's in fact not in GCC's
-// libstdc++-v3).  Copying the UTF-8 data into a temporary buffer isn't an
-// option since this operation is quite time critical.  The implementation
-// is quite different from g_utf8_get_char() -- both more generic and faster.
-//
-// By looking at the first byte of a UTF-8 character one can determine the
-// number of bytes used.  GLib offers the g_utf8_skip[] array for this purpose,
-// but accessing this global variable would introduce a function call to fetch
-// the Global Offset Table, plus two levels of indirection in order to read the
-// value.  Even worse, fetching the GOT is always done right at the start of
-// the function instead of the branch that actually uses the variable.
-//
-// Fortunately, there's a better way to get the byte count.  As this table
-// shows, there's a nice regular pattern in the UTF-8 encoding scheme:
-//
-// 0x00000000 - 0x0000007F: 0xxxxxxx
-// 0x00000080 - 0x000007FF: 110xxxxx 10xxxxxx
-// 0x00000800 - 0x0000FFFF: 1110xxxx 10xxxxxx 10xxxxxx
-// 0x00010000 - 0x001FFFFF: 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
-// 0x00200000 - 0x03FFFFFF: 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
-// 0x04000000 - 0x7FFFFFFF: 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
-//
-// Except for the single byte case, the number of leading 1-bits equals the
-// byte count.  All that is needed is to shift the first byte to the left
-// until bit 7 becomes 0.  Naturally, doing so requires a loop -- but since
-// we already have one, no additional cost is introduced.  This shifting can
-// further be combined with the computation of the bitmask needed to eliminate
-// the leading length bits, thus saving yet another register.
-//
+/*
+ * We need our own version of g_utf8_get_char(), because the std::string
+ * iterator is not necessarily a plain pointer (it's in fact not in GCC's
+ * libstdc++-v3).  Copying the UTF-8 data into a temporary buffer isn't an
+ * option since this operation is quite time critical.  The implementation
+ * is quite different from g_utf8_get_char() -- both more generic and likely
+ * faster.
+ *
+ * By looking at the first byte of a UTF-8 character one can determine the
+ * number of bytes used.  GLib offers the g_utf8_skip[] array for this purpose,
+ * but accessing this global variable would, on IA32 at least, introduce
+ * a function call to fetch the Global Offset Table, plus two levels of
+ * indirection in order to read the value.  Even worse, fetching the GOT is
+ * always done right at the start of the function instead of the branch that
+ * actually uses the variable.
+ *
+ * Fortunately, there's a better way to get the byte count.  As this table
+ * shows, there's a nice regular pattern in the UTF-8 encoding scheme:
+ *
+ * 0x00000000 - 0x0000007F: 0xxxxxxx
+ * 0x00000080 - 0x000007FF: 110xxxxx 10xxxxxx
+ * 0x00000800 - 0x0000FFFF: 1110xxxx 10xxxxxx 10xxxxxx
+ * 0x00010000 - 0x001FFFFF: 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
+ * 0x00200000 - 0x03FFFFFF: 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
+ * 0x04000000 - 0x7FFFFFFF: 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
+ *
+ * Except for the single byte case, the number of leading 1-bits equals the
+ * byte count.  All that is needed is to shift the first byte to the left
+ * until bit 7 becomes 0.  Naturally, doing so requires a loop -- but since
+ * we already have one, no additional cost is introduced.  This shifting can
+ * further be combined with the computation of the bitmask needed to eliminate
+ * the leading length bits, thus saving yet another register.
+ *
+ * Note:  If you change this code, it is advisable to also review what the
+ * compiler makes of it in the assembler output.  Except for some pointless
+ * register moves, the generated code is sufficiently close to the optimum
+ * with GCC 4.1.2 on x86_64.
+ */
 gunichar get_unichar_from_std_iterator(std::string::const_iterator pos)
 {
   unsigned int result = static_cast<unsigned char>(*pos);
@@ -258,21 +273,16 @@ gunichar get_unichar_from_std_iterator(std::string::const_iterator pos)
   {
     unsigned int mask = 0x40;
 
-    // This loop may look somewhat strange, but this happens to be the
-    // one variant g++ 3.3 generates optimum code for.  (In particular,
-    // the unconditional break avoids counterproductive loop alignment.)
-    for(++pos;; ++pos)
+    do
     {
-      mask <<= 5;
-      const unsigned int c = static_cast<unsigned char>(*pos);
-      result = (result << 6) + c - 0x80;
-
-      if((result & mask) != 0)
-        continue;
-      break;
+      result <<= 6;
+      const unsigned int c = static_cast<unsigned char>(*++pos);
+      mask   <<= 5;
+      result  += c - 0x80;
     }
+    while((result & mask) != 0);
 
-    result &= --mask;
+    result &= mask - 1;
   }
 
   return result;
@@ -673,7 +683,10 @@ ustring& ustring::erase()
 
 ustring::iterator ustring::erase(ustring::iterator p)
 {
-  return iterator(string_.erase(p.base()));
+  ustring::iterator iter_end = p;
+  ++iter_end;
+
+  return iterator(string_.erase(p.base(), iter_end.base()));
 }
 
 ustring::iterator ustring::erase(ustring::iterator pbegin, ustring::iterator pend)
@@ -1125,7 +1138,7 @@ bool ustring::is_ascii() const
 
   for(; p != pend; ++p)
   {
-    if((*p & '\x80') != 0)
+    if((static_cast<unsigned char>(*p) & 0x80u) != 0)
       return false;
   }
 
@@ -1171,6 +1184,58 @@ std::string ustring::casefold_collate_key() const
   return std::string(ScopedPtr<char>(key_buf).get());
 }
 
+/**** Glib::ustring -- Message formatting **********************************/
+
+// static
+ustring ustring::compose_argv(const Glib::ustring& fmt, int argc, const ustring* const* argv)
+{
+  std::string::size_type result_size = fmt.raw().size();
+
+  // Guesstimate the final string size.
+  for (int i = 0; i < argc; ++i)
+    result_size += argv[i]->raw().size();
+
+  std::string result;
+  result.reserve(result_size);
+
+  const char* const pfmt = fmt.raw().c_str();
+  const char* start = pfmt;
+
+  while (const char* const stop = std::strchr(start, '%'))
+  {
+    if (stop[1] == '%')
+    {
+      result.append(start, stop - start + 1);
+      start = stop + 2;
+    }
+    else
+    {
+      const int index = Ascii::digit_value(stop[1]) - 1;
+
+      if (index >= 0 && index < argc)
+      {
+        result.append(start, stop - start);
+        result += argv[index]->raw();
+        start = stop + 2;
+      }
+      else
+      {
+        const char* const next = (stop[1] != '\0') ? g_utf8_next_char(stop + 1) : (stop + 1);
+
+        // Copy invalid substitutions literally to the output.
+        result.append(start, next - start);
+
+        g_warning("invalid substitution \"%s\" in fmt string \"%s\"",
+                  result.c_str() + result.size() - (next - stop), pfmt);
+        start = next;
+      }
+    }
+  }
+
+  result.append(start, pfmt + fmt.raw().size() - start);
+
+  return result;
+}
 
 /**** Glib::ustring::SequenceToString **************************************/
 
@@ -1186,34 +1251,201 @@ ustring::SequenceToString<Glib::ustring::const_iterator,gunichar>
   std::string(pbegin.base(), pend.base())
 {}
 
+/**** Glib::ustring::FormatStream ******************************************/
+
+ustring::FormatStream::FormatStream()
+:
+  stream_ ()
+{}
+
+ustring::FormatStream::~FormatStream()
+{}
+
+ustring ustring::FormatStream::to_string() const
+{
+  GError* error = 0;
+
+#ifdef GLIBMM_HAVE_WIDE_STREAM
+  const std::wstring str = stream_.str();
+
+# if defined(__STDC_ISO_10646__) && SIZEOF_WCHAR_T == 4
+  // Avoid going through iconv if wchar_t always contains UCS-4.
+  glong n_bytes = 0;
+  const ScopedPtr<char> buf (g_ucs4_to_utf8(reinterpret_cast<const gunichar*>(str.data()),
+                                            str.size(), 0, &n_bytes, &error));
+# elif defined(G_OS_WIN32) && SIZEOF_WCHAR_T == 2
+  // Avoid going through iconv if wchar_t always contains UTF-16.
+  glong n_bytes = 0;
+  const ScopedPtr<char> buf (g_utf16_to_utf8(reinterpret_cast<const gunichar2*>(str.data()),
+                                             str.size(), 0, &n_bytes, &error));
+# else
+  gsize n_bytes = 0;
+  const ScopedPtr<char> buf (g_convert(reinterpret_cast<const char*>(str.data()),
+                                       str.size() * sizeof(std::wstring::value_type),
+                                       "UTF-8", "WCHAR_T", 0, &n_bytes, &error));
+# endif /* !(__STDC_ISO_10646__ || G_OS_WIN32) */
+
+#else /* !GLIBMM_HAVE_WIDE_STREAM */
+  const std::string str = stream_.str();
+
+  gsize n_bytes = 0;
+  const ScopedPtr<char> buf (g_locale_to_utf8(str.data(), str.size(), 0, &n_bytes, &error));
+#endif /* !GLIBMM_HAVE_WIDE_STREAM */
+
+  if (error)
+  {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
+    Glib::Error::throw_exception(error);
+#else
+    g_warning("%s: %s", G_STRFUNC, error->message);
+    g_error_free(error);
+    return ustring();
+#endif
+  }
+
+  return ustring(buf.get(), buf.get() + n_bytes);
+}
 
 /**** Glib::ustring -- stream I/O operators ********************************/
 
 std::istream& operator>>(std::istream& is, Glib::ustring& utf8_string)
 {
-  std::string locale_string;
-  is >> locale_string;
+  std::string str;
+  is >> str;
+
+  GError* error = 0;
+  gsize n_bytes = 0;
+  const ScopedPtr<char> buf (g_locale_to_utf8(str.data(), str.size(), 0, &n_bytes, &error));
+
+  if (error)
+  {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
+    Glib::Error::throw_exception(error);
+#else
+    g_warning("%s: %s", G_STRFUNC, error->message);
+    g_error_free(error);
+    return is;
+#endif
+  }
+
+  utf8_string.assign(buf.get(), buf.get() + n_bytes);
 
-  #ifdef GLIBMM_EXCEPTIONS_ENABLED
-  utf8_string = Glib::locale_to_utf8(locale_string);
-  #else
-  std::auto_ptr<Glib::Error> error; //TODO: Check this? 
-  utf8_string = Glib::locale_to_utf8(locale_string, error);
-  #endif //GLIBMM_EXCEPTIONS_ENABLED
   return is;
 }
 
 std::ostream& operator<<(std::ostream& os, const Glib::ustring& utf8_string)
 {
-  #ifdef GLIBMM_EXCEPTIONS_ENABLED
-  os << Glib::locale_from_utf8(utf8_string);
-  #else
-  std::auto_ptr<Glib::Error> error; //TODO: Check this? 
-  os << Glib::locale_from_utf8(utf8_string, error);
-  #endif //GLIBMM_EXCEPTIONS_ENABLED
+  GError* error = 0;
+  const ScopedPtr<char> buf (g_locale_from_utf8(utf8_string.raw().data(),
+                                                utf8_string.raw().size(), 0, 0, &error));
+  if (error)
+  {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
+    Glib::Error::throw_exception(error);
+#else
+    g_warning("%s: %s", G_STRFUNC, error->message);
+    g_error_free(error);
+    return os;
+#endif
+  }
+
+  // This won't work if the string contains NUL characters.  Unfortunately,
+  // std::ostream::write() ignores format flags, so we cannot use that.
+  // The only option would be to create a temporary std::string.  However,
+  // even then GCC's libstdc++-v3 prints only the characters up to the first
+  // NUL.  Given this, there doesn't seem much of a point in allowing NUL in
+  // formatted output.  The semantics would be unclear anyway: what's the
+  // screen width of a NUL?
+  os << buf.get();
 
   return os;
 }
 
-} // namespace Glib
+#ifdef GLIBMM_HAVE_WIDE_STREAM
 
+std::wistream& operator>>(std::wistream& is, ustring& utf8_string)
+{
+  GError* error = 0;
+
+  std::wstring wstr;
+  is >> wstr;
+
+#if defined(__STDC_ISO_10646__) && SIZEOF_WCHAR_T == 4
+  // Avoid going through iconv if wchar_t always contains UCS-4.
+  glong n_bytes = 0;
+  const ScopedPtr<char> buf (g_ucs4_to_utf8(reinterpret_cast<const gunichar*>(wstr.data()),
+                                            wstr.size(), 0, &n_bytes, &error));
+#elif defined(G_OS_WIN32) && SIZEOF_WCHAR_T == 2
+  // Avoid going through iconv if wchar_t always contains UTF-16.
+  glong n_bytes = 0;
+  const ScopedPtr<char> buf (g_utf16_to_utf8(reinterpret_cast<const gunichar2*>(wstr.data()),
+                                             wstr.size(), 0, &n_bytes, &error));
+#else
+  gsize n_bytes = 0;
+  const ScopedPtr<char> buf (g_convert(reinterpret_cast<const char*>(wstr.data()),
+                                       wstr.size() * sizeof(std::wstring::value_type),
+                                       "UTF-8", "WCHAR_T", 0, &n_bytes, &error));
+#endif /* !(__STDC_ISO_10646__ || G_OS_WIN32) */
+
+  if (error)
+  {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
+    Glib::Error::throw_exception(error);
+#else
+    g_warning("%s: %s", G_STRFUNC, error->message);
+    g_error_free(error);
+    return is;
+#endif
+  }
+
+  utf8_string.assign(buf.get(), buf.get() + n_bytes);
+
+  return is;
+}
+
+std::wostream& operator<<(std::wostream& os, const ustring& utf8_string)
+{
+  GError* error = 0;
+
+#if defined(__STDC_ISO_10646__) && SIZEOF_WCHAR_T == 4
+  // Avoid going through iconv if wchar_t always contains UCS-4.
+  const ScopedPtr<gunichar> buf (g_utf8_to_ucs4(utf8_string.raw().data(),
+                                                utf8_string.raw().size(), 0, 0, &error));
+#elif defined(G_OS_WIN32) && SIZEOF_WCHAR_T == 2
+  // Avoid going through iconv if wchar_t always contains UTF-16.
+  const ScopedPtr<gunichar2> buf (g_utf8_to_utf16(utf8_string.raw().data(),
+                                                  utf8_string.raw().size(), 0, 0, &error));
+#else
+  // TODO: For some reason the conversion from UTF-8 to WCHAR_T doesn't work
+  // with g_convert(), while iconv on the command line handles it just fine.
+  // Maybe a bug in GLib?
+  const ScopedPtr<char> buf (g_convert(utf8_string.raw().data(), utf8_string.raw().size(),
+                                       "WCHAR_T", "UTF-8", 0, 0, &error));
+#endif /* !(__STDC_ISO_10646__ || G_OS_WIN32) */
+
+  if (error)
+  {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
+    Glib::Error::throw_exception(error);
+#else
+    g_warning("%s: %s", G_STRFUNC, error->message);
+    g_error_free(error);
+    return os;
+#endif
+  }
+
+  // This won't work if the string contains NUL characters.  Unfortunately,
+  // std::wostream::write() ignores format flags, so we cannot use that.
+  // The only option would be to create a temporary std::wstring.  However,
+  // even then GCC's libstdc++-v3 prints only the characters up to the first
+  // NUL.  Given this, there doesn't seem much of a point in allowing NUL in
+  // formatted output.  The semantics would be unclear anyway: what's the
+  // screen width of a NUL?
+  os << reinterpret_cast<wchar_t*>(buf.get());
+
+  return os;
+}
+
+#endif /* GLIBMM_HAVE_WIDE_STREAM */
+
+} // namespace Glib
